/* This is an implementation of shift scheduling problem
 * for demonstration CGI by S. Moriguchi.
 *
 *  c.f. G. Koole and E. van der Sluis (2003): Optimal shift scheduling
 *  with a global service level constraint, IIE Transactions, Vol. 35,
 *  pp. 1049-1055.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "odicon.h"


#define INTERVAL 13
#define DELTA 0.000005

/* determined in demo_shift_schedule() */
static double c = 0.01, mu = 10.0;
static double lambda[INTERVAL] = {88, 87, 86, 85, 55, 1, 1, 1, 1, 1, 1, 1, 1};
static int n_total;

/* x^n / n! */
double exp_per_fact(double x, int n) {
    int i;
    double r = 1.0;

    for (i=n; i>=1; i--) r = r * x / i;
    return r;
}

/* pi = b / (a + b), safe even if b=inf */
double pi(double a, double b) {
    if (b < a) {
        double s = b / a;
	return s / ( 1 + s );
    } else {
        double t = a / b;
	return 1 / ( t + 1 );
    }
}


double gi(int s, int i) {
    double rho=0, beta=0, alpha=0, rho_n=0, gi;
    int j;
    
    rho = lambda[i-1] / mu;
    rho_n = rho / s;
    
    beta = exp_per_fact(rho,s) / (1-rho_n);
    
    for(j=0; j<s; j++){
        alpha += exp_per_fact(rho,j);
    }
    gi = lambda[i-1] * pi(alpha, beta) * exp(-s * mu*(1-rho_n)*c);

    printf(" gi[%d] = %g,\talpha = %g,\tbeta = %g,\tpi = %g,\texp(*) = %g\n",
	   i, gi, alpha, beta, pi(alpha, beta), exp(-s * mu*(1-rho_n)*c));
    return gi;
}


/* f1(x) = \sum g_i */
static double f1(int dim, int x[]) {
    double r = 0, LAMBDA;
    int j;
  
    assert(dim == 3);  /* check dimension */

    if (x[0] <= 0) return PLUS_INFINITY;
    for (j=1; j<dim; j++) {
        if (x[j] <= 0 || x[j]<=x[j-1]) return PLUS_INFINITY;
    }
    for (j=0; j<dim; j++) {
      assert(lambda[j] < n_total * mu);
    }

    r +=  gi(x[0]-0, 1);
    r +=  gi(x[0]-0, 2);
    r +=  gi(x[1]-0, 3);
    r +=  gi(x[1]-0, 4);
    r +=  gi(x[1]-0, 5);
    r +=  gi(x[2]-x[0], 6);
    r +=  gi(x[2]-x[0], 7);
    r +=  gi(x[2]-x[1], 8);
    r +=  gi(n_total-x[1], 9);
    r +=  gi(n_total-x[1], 10);
    r +=  gi(n_total-x[2], 11);
    r +=  gi(n_total-x[2], 12);
    r +=  gi(n_total-x[2], 13);

    LAMBDA = 0;
    for(j=0; j<INTERVAL; j++){
        LAMBDA += lambda[j];
    }
    
    return r / LAMBDA;
}

#define DIM 3

void demo_shift_schedule(const int dim) {
    int y[DIM+1] = {1, 1, 1, 7};  /* initial solution */
    int x[DIM+1];
    double ans;
    int i, sum, last;

    assert( dim == DIM );
    /* for (i=0; i<=10; i++) printf("%d! = %f\n", i, factorial(i)); */
    /* printf("gi = %f\n", gi(5,5)); */

    fvector_scan(1, &c);
    fvector_scan(1, &mu);
    fvector_scan(INTERVAL, lambda);
    ivector_scan(DIM+1, y);

    sum = 0;
    for (i=0; i<DIM+1; i++) {
        sum += y[i];
        x[i] = sum;
    }
    n_total = x[DIM];
    ans = lgconv_minimize(dim, f1, x);
    /*printf("S(x) = %f\n", -ans);*/
    printf("N = %d\n", n_total);
    last = 0;
    for (i=0; i<DIM+1; i++) {
        y[i] = x[i] - last;
        last = x[i];
    }
    printf("\n");
    for(i=0; i<DIM+1; i++) {
        printf("x[%d] = %d\n", i+1, x[i]);
    }
    printf("\n");
    for(i=0; i<DIM+1; i++) {
        printf("y[%d] = %d\n", i+1, y[i]);
    }
}
